#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <errno.h>

#include "nfspd_ctl.h"
#include "nfspd_stub.h"

#define ERR(fmt,args...) do { int tmp = errno; fprintf(stderr, "libnfspdtcl:" fmt "\n", ##args); errno = tmp; } while(0)

static int ctl_sock = -1;

int nfspd_initialize() {
	struct sockaddr_un sun;
	char *home;

	ctl_sock = socket(PF_UNIX, SOCK_STREAM, 0);
	if (ctl_sock < 0) {
		ERR("stub socket() failed errno=%d\n", errno);
		goto error;
	}
	
	home = getenv(SOCK_CTL_ENV_VAR);
	if (!home) {
		ERR("%s must be set", SOCK_CTL_ENV_VAR);
		goto error_cleanup;
	}

	sun.sun_family = AF_UNIX;
	if (sizeof(sun.sun_path) < strlen(home) + 1 + strlen(SOCK_CTL_NAME) + 1) {
		ERR("path to Unix control socket too long (>%d)", sizeof(sun.sun_path));
		goto error_cleanup;
	}

	strcpy(sun.sun_path, home);
	strcat(sun.sun_path, "/");
	strcat(sun.sun_path, SOCK_CTL_NAME);

	if (-1 == connect(ctl_sock, (struct sockaddr *) &sun, sizeof(sun))) {
		ERR("stub connect() failed errno=%d\n", errno);
		goto error_cleanup;
	}

	return 0;

error_cleanup:
	(void) close(ctl_sock);
error:
	ctl_sock = -1;
	return -1;
}


int nfspd_terminate() {
	int ret;
	
	ret = close(ctl_sock);

	return ret;
}


static int send_mess(const struct message_t * const messp) {
	int ret;
	int pos;

	if (ctl_sock < 0) {
		errno = ENOTCONN;
		return -1;
	}

	for (pos = 0; pos < sizeof(*messp); ) {
		int tmp;
		tmp = write(ctl_sock, ((char *)messp) + pos , sizeof(*messp) - pos);
		if (tmp == -1) {
			ERR("write error errno=%d", errno);
			errno = ECONNABORTED;
			return -1;
		}
		pos += tmp;
	}
	for (pos = 0; pos < sizeof(ret); ) {
	       int tmp;
       		tmp = read(ctl_sock, ((char *)&ret) + pos, sizeof(ret) - pos);
		if (tmp == -1) {
			ERR("read error errno=%d", errno);
			errno = ECONNABORTED;
			return -1;
		}
		pos += tmp;
	}

	if (ret < 0) {
		errno = -ret;
		return -1;
	}

	return ret;
}


/* creates a new iod descriptor given its address and port */
int nfspd_addiod(const struct sockaddr_in * const addr) {
	struct message_t mess;
	
	mess.magic = NFSP_MAGIC;
	mess.fnct = NFSP_ADD_IOD;
	mess.addr = *addr;

	return send_mess(&mess);
}


/* free an iod descriptor */
int nfspd_freeiod(const int ih) {
	struct message_t mess;
	
	mess.magic = NFSP_MAGIC;
	mess.fnct = NFSP_FREE_IOD;
	mess.ih = ih;

	return send_mess(&mess);
}


/* bind an iod to a viod */
int nfspd_bindiod(int iviod, int ih) {
	struct message_t mess;
	
	mess.magic = NFSP_MAGIC;
	mess.fnct = NFSP_BIND;
	mess.iviod = iviod;
	mess.ih = ih;

	return send_mess(&mess);
}


int nfspd_unbindiod(int ih) {
	struct message_t mess;
	
	mess.magic = NFSP_MAGIC;
	mess.fnct = NFSP_UNBIND;
	mess.ih = ih;

	return send_mess(&mess);
}


int nfspd_setnbviods(int nb) {
	struct message_t mess;
	
	mess.magic = NFSP_MAGIC;
	mess.fnct = NFSP_SETNBVIODS;
	mess.nbviods = nb;
	
	return send_mess(&mess);
}


int nfspd_getnbviods() {
	struct message_t mess;

	mess.magic = NFSP_MAGIC;
	mess.fnct = NFSP_GETNBVIODS;

	return send_mess(&mess);
}

int nfspd_drop() {
	struct message_t mess;

	mess.magic = NFSP_MAGIC;
	mess.fnct = NFSP_DROP;

	return send_mess(&mess);
}

int nfspd_forward() {
	struct message_t mess;

	mess.magic = NFSP_MAGIC;
	mess.fnct = NFSP_FORWARD;

	return send_mess(&mess);
}

int nfspd_flush() {
	struct message_t mess;

	mess.magic = NFSP_MAGIC;
	mess.fnct = NFSP_FLUSH;

	return send_mess(&mess);
}

int nfspd_get_viod_size(const int iviod) {
	struct message_t mess;

	mess.magic = NFSP_MAGIC;
	mess.fnct = NFSP_GET_VIOD_SIZE;
	mess.iviod = iviod;

	return send_mess(&mess);
}


