// nfs_dispatch.c
// This file contains the function dispatch table and most of the code to
// handle stripping among iod's
//
// Changelog:
//
// 200208xx: Adrien Lebre <adrien.lebre@imag.fr>
//   Add Write mirror request
//   
// 200205xx: Adrien Lebre <adrien.lebre@imag.fr>
//   IPtoMAC() function
// 
// 2001xxxx: Pierre Lombard <pierre.lombard@imag.fr>
//   initial changes (READ/WRITE) for NFS support
// 
// ***************************************************************************
// Authors:
//	Donald J. Becker, <becker@super.org>
//	Rick Sladkey, <jrs@world.std.com>
//	Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
//	Olaf Kirch, <okir@monad.swb.de>
// 
// 	This software may be used for any purpose provided the above copyright
// 	notice is retained.  It is supplied as is, with no warranty expressed or
// 	implied.
// ***************************************************************************

#include "nfsd.h"
#include "nfs_prot.h"
#include "rpcmisc.h"

#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "iodng.h"
#include <sys/types.h>
#include <arpa/inet.h>
#include <unistd.h>
//#include "dbg.h"

/* Required for struct arpreq in IPtoMAC() - man 7 arp */
#include <net/if_arp.h>

#ifdef VIODS_SUPPORT
#include "nfsp_viods.h"
#else /* VIODS_SUPPORT */
#include "nfsp_iods.h"
#endif /* !VIODS_SUPPORT */

#include "iodng.h"
#include "unlinkd.h"

/* Socket used to send UDP messages to iods - defined in nfsd.*/
extern int iods_udp_fd;

/* This functions wraps the way to access to the xid in the transp structure.
 * Granted it's a piggy way to get this but xid number is hidden in a private
 * sunrpc data in glibc... */
static inline u_long get_xid(SVCXPRT *transp) {
	return htonl(((unsigned long *)(transp->xp_p2))[1]);
}


/*
 * These are the global variables that hold all argument and result data.
 */
union argument_types 	argument;
union result_types	result;

/*
 * The time at which we received the request.
 * Useful for various book-keeping things
 */
time_t			nfs_dispatch_time;

/*
 * This is a dispatch table to simplify error checking,
 * and supply return attributes for NFS functions.
 */

#ifdef __STDC__
#define CONCAT(a,b)	a##b
#define CONCAT3(a,b,c)	a##b##c
#define STRING(a)	#a
#else /* __STDC__ */
#define CONCAT(a,b)	a/**/b
#define CONCAT3(a,b,c)	a/**/b/**/c
#define STRING(a)	"a"
#endif /* !__STDC__ */

#define table_ent(res_type, arg_type, funct) {		\
	sizeof(res_type), sizeof(arg_type),		\
	(xdrproc_t) CONCAT(xdr_,res_type),		\
	(xdrproc_t) CONCAT(xdr_,arg_type),		\
	(int (*)()) CONCAT3(nfsd_nfsproc_,funct,_2),	\
	STRING(funct), CONCAT(pr_,arg_type)		\
}

/* This definition makes the first table entry use xdr_char instead of
 * xdr_void. The NULL procedure call over TCP fails for this reason
#define nil	char
 */
#define nil	void
#define xdr_nil	xdr_void
#define pr_nil	pr_void
#define pr_char	pr_void

struct dispatch_entry {
    int		res_size, arg_size;		/* sizeof the res/arg structs	*/
    xdrproc_t	xdr_result;
    xdrproc_t	xdr_argument;
    int		(*funct)();			/* function handler		*/
    char	*name;				/* name of function		*/
    char	*(*log_print)();		/* ptr to debug handler		*/
};

static char *pr_void(void);
static char *pr_nfs_fh(nfs_fh *argp);
static char *pr_sattrargs(sattrargs *argp);
static char *pr_diropargs(diropargs *argp);
static char *pr_readargs(readargs *argp);
static char *pr_writeargs(writeargs *argp);
static char *pr_createargs(createargs *argp);
static char *pr_renameargs(renameargs *argp);
static char *pr_linkargs(linkargs *argp);
static char *pr_symlinkargs(symlinkargs *argp);
static char *pr_readdirargs(readdirargs *argp);

static struct dispatch_entry dtable[] = {
	table_ent(nil,nil,null),			/* NULL */
	table_ent(attrstat,nfs_fh,getattr),		/* GETATTR */
	table_ent(attrstat,sattrargs,setattr),		/* SETATTR */
	table_ent(nil,nil,root),			/* ROOT */
	table_ent(diropres,diropargs,lookup),		/* LOOKUP */
	table_ent(readlinkres,nfs_fh,readlink),		/* READLINK */
	table_ent(readres,readargs,read),		/* READ */
	table_ent(nil,nil,writecache),			/* WRITECACHE */
	table_ent(attrstat,writeargs,write),		/* WRITE */
	table_ent(diropres,createargs,create),		/* CREATE */
	table_ent(nfsstat,diropargs,remove),		/* REMOVE */
	table_ent(nfsstat,renameargs,rename),		/* RENAME */
	table_ent(nfsstat,linkargs,link),		/* LINK */
	table_ent(nfsstat,symlinkargs,symlink),		/* SYMLINK */
	table_ent(diropres,createargs,mkdir),		/* MKDIR */
	table_ent(nfsstat,diropargs,rmdir),		/* RMDIR */
	table_ent(readdirres,readdirargs,readdir),	/* READDIR */
	table_ent(statfsres,nfs_fh,statfs),		/* STATFS */
};


// This function finds the MAC address matching the IP addrIP by looking into
// the arp cache.  It uses ioctl on the UDP socket iods_udp_fd to do this.
//
// WARNING: only use with AF_INET socket family
//
// IN : IP the ip address in network format.
// OUT : MAC address (unsigned char[6])
//
// TODO find a method to dynamically assign the interface
// TODO only run with the MAC ADDRESS size of 6
static int IPtoMAC (const in_addr_t addrIP, char * MAC) {
	int tmpSocket = iods_udp_fd;
	struct arpreq arpRequest;
	struct sockaddr_in *sin;

	memset(&arpRequest, '\0', sizeof(arpRequest));
	strcpy(arpRequest.arp_dev, "eth0");
	sin = (struct sockaddr_in *) &(arpRequest.arp_pa);
	sin->sin_family = AF_INET;
	arpRequest.arp_flags = ATF_COM;
	memcpy(&(sin->sin_addr.s_addr), &addrIP, sizeof(in_addr_t));

	if (ioctl(tmpSocket, SIOCGARP, &arpRequest)) {
		Dprintf(L_WARNING, "ioctl with SIOCGARP failed errno=%d", errno);
		memset(MAC, '\0', 6);
	        return -1;
	}

       	// A MAC address is 6 bytes
	memcpy(MAC, arpRequest.arp_ha.sa_data, 6);
	
	return 0;
}


// May be find_iod(u_int const inode, int const seed, int const offset)
static inline int find_iod(int const seed, int const offset) {
	unsigned int hash;

	// Better hash functions may be found on:
	// http://bitmover.com:8888//home/bk/bugfixes/anno/src/mdbm/hash.c@1.3?nav=src/src/mdbm
	hash = seed + (offset/BLOCKSIZE);

	return (hash % viods_nb);
}


static int forward_read(SVCXPRT *transp, xdrproc_t outproc, char *out) {
	// READ req
	msg_iod_t m;
	struct sockaddr_in *dasock = svc_getcaller(transp);
	u_int offset = argument.nfsproc_read_2_arg.offset;
	u_int datalen = argument.nfsproc_read_2_arg.count;
	int iod_no, ret;

	Dprintf(D_NFSP_DISPATCH, "forward_read: offset=%d datalen=%d",offset,datalen);

	m.client_addr = dasock->sin_addr.s_addr;
	m.client_port = dasock->sin_port;

	if (IPtoMAC(m.client_addr, m.client_mac_addr)) {
		Dprintf(L_WARNING, "forward_read: IPtoMAC failed");
		return -1;
	}

	m.server_addr = 0;			// iod will fill it
	m.server_port = htons(NFSport);

	// Does the request spans onto several BLOCKS ?
	if ((offset % BLOCKSIZE) + datalen <= BLOCKSIZE) {
		// Simple case: NO
		Dprintf(D_NFSP_DISPATCH, "forward_read: READREQ o=0x%8.8x bs=0x%8.8x datalen=0x%8.8x",
			offset, BLOCKSIZE, datalen);

		m.magic = NFSP_MAGIC_MSG;
		m.msgtype = NFSP_READ_REQ;

		m.readreq.xid = get_xid(transp);

		m.readreq.stripe_bs = BLOCKSIZE;
		m.readreq.stripe_w = viods_nb;

		m.readreq.offset = offset;
		m.readreq.datalen = datalen;

		// We can't use result.readres.readres_u.reply.attributes.fileid since
		// it's not a real inode number but the value returned to the user if
		// he asks the inode value
		m.readreq.inode = global_real_inode;
		m.readreq.seed = global_seed;

		memcpy(&m.readreq.attributes,
			&result.readres.readres_u.reply.attributes,
			sizeof(result.readres.readres_u.reply.attributes));
	
		iod_no = find_iod(m.readreq.seed, m.readreq.offset);
		ret = sendto(iods_udp_fd, (char *)&m, MSG_IOD_READ_LEN, 0,
			get_iod(iod_no), sizeof(struct sockaddr_in));

	} else {
		// Trickier case: YES
		int iod_no_next;

		Dprintf(D_NFSP_DISPATCH, "forward_read: FWDREAD_1_REQ o=0x%8.8x bs=0x%8.8x datalen=0x%8.8x",
			offset, BLOCKSIZE, datalen);

		m.magic = NFSP_MAGIC_MSG;
		m.msgtype = NFSP_FWDREAD_1_REQ;
		m.fwdreadreq.xid = get_xid(transp);

		m.fwdreadreq.stripe_bs = BLOCKSIZE;
		m.fwdreadreq.stripe_w = viods_nb;

		// result.readres.readres_u.reply.attributes.fileid is not
		// usable as it's not a real inode number but the value
		// returned to the user when the inode value is asked
		m.fwdreadreq.inode = global_real_inode;
		m.fwdreadreq.seed = global_seed;

		memcpy(&m.fwdreadreq.attributes,
			&result.readres.readres_u.reply.attributes,
			sizeof(result.readres.readres_u.reply.attributes));

		m.fwdreadreq.offset = offset;
		m.fwdreadreq.datalen = datalen;

		iod_no = find_iod(m.fwdreadreq.seed, m.fwdreadreq.offset);
		iod_no_next = (iod_no + 1) % viods_nb;

		memcpy(&m.fwdreadreq.nextiod,
			get_iod(iod_no_next),
			sizeof(struct sockaddr_in));

		m.fwdreadreq.nextlen = 0x6666;

		ret = sendto(iods_udp_fd, (char *)&m, MSG_IOD_FWDREAD_LEN(0), 0,
				get_iod(iod_no), sizeof(struct sockaddr_in));
	}
	return ret;
}

//#ifndef WRITEMIRROR_SUPPORT
#if 1
static int forward_write(SVCXPRT *transp, xdrproc_t outproc, char *out) {
	// WRITE req
	msg_iod_t m;
	struct sockaddr_in *dasock = svc_getcaller(transp);
	u_int offset = argument.nfsproc_write_2_arg.offset;
	u_int datalen = argument.nfsproc_write_2_arg.data.data_len;
	int iod_no, ret;

	Dprintf(D_NFSP_DISPATCH, "forward_write: offset=%d datalen=%d", offset, datalen);

	m.client_addr = dasock->sin_addr.s_addr;
	m.client_port = dasock->sin_port;

	if (IPtoMAC(m.client_addr, m.client_mac_addr)) {
		Dprintf(L_WARNING, "forward_write: IPtoMAC failed");
		return -1;
	}
	
	m.server_addr = 0;		// iod will fill it
	m.server_port = htons(NFSport);

	// Does the request spans onto several BLOCKS ?
	if ((offset % BLOCKSIZE) + datalen <= BLOCKSIZE) {
		// Simple case: NO
		Dprintf(D_NFSP_DISPATCH, "forward_write: WRITEREQ o=0x%8.8x bs=0x%8.8x datalen=0x%8.8x",
			offset, BLOCKSIZE, datalen);

		m.magic = NFSP_MAGIC_MSG;
		m.msgtype = NFSP_WRITE_REQ;

		m.writereq.xid = get_xid(transp);

		m.writereq.stripe_bs = BLOCKSIZE;
		m.writereq.stripe_w = viods_nb;
	
		m.writereq.offset = offset;
		m.writereq.datalen = datalen;

		m.writereq.inode = global_real_inode;
		m.writereq.seed = global_seed;

		memcpy(&m.writereq.attributes, &result.attrstat.attrstat_u.attributes,
			sizeof(result.attrstat.attrstat_u.attributes));

		memcpy(m.writereq.dataval,
			argument.nfsproc_write_2_arg.data.data_val, datalen);

		Dprintf(D_NFSP_DISPATCH, "forward_write data_len=%d rev=%d",
			argument.nfsproc_write_2_arg.data.data_len,
			ntohl(argument.nfsproc_write_2_arg.data.data_len));

		iod_no = find_iod(m.writereq.seed, m.writereq.offset);
		ret = sendto(iods_udp_fd, (char *)&m, MSG_IOD_WRITE_LEN(datalen), 0,
			get_iod(iod_no), sizeof(struct sockaddr_in));
	} else {
		// Trickier case: YES
		int iod_no_next;

		Dprintf(D_NFSP_DISPATCH, "forward_write: FWDREAD_1_WRITE o=0x%8.8x bs=0x%8.8x datalen=0x%8.8x",
			offset, BLOCKSIZE, datalen);

		m.magic = NFSP_MAGIC_MSG;
		m.msgtype = NFSP_FWDWRITE_1_REQ;
		m.fwdwritereq.xid = get_xid(transp);

		m.fwdwritereq.stripe_bs = BLOCKSIZE;
		m.fwdwritereq.stripe_w = viods_nb;

		m.fwdwritereq.inode = global_real_inode;
		m.fwdwritereq.seed = global_seed;

		memcpy(&m.writereq.attributes, &result.attrstat.attrstat_u.attributes,
			sizeof(result.attrstat.attrstat_u.attributes));

		m.fwdwritereq.offset = offset;
		m.fwdwritereq.datalen = datalen;
		memcpy(m.fwdwritereq.dataval,
			argument.nfsproc_write_2_arg.data.data_val, datalen);

		iod_no = find_iod(m.fwdreadreq.seed, m.fwdreadreq.offset);
		iod_no_next = (iod_no + 1) % viods_nb;

		memcpy(&m.fwdwritereq.nextiod,
			get_iod(iod_no_next),
			sizeof(struct sockaddr_in));

		
		m.fwdwritereq.nextlen = 0x6666;

		ret = sendto(iods_udp_fd, (char *)&m, MSG_IOD_FWDWRITE_LEN(datalen), 0,
			get_iod(iod_no), sizeof(struct sockaddr_in));
	}
	return ret;

}

#else /* !WRITEMIRROR_SUPPORT */

static int forward_write_mirror(SVCXPRT *transp, xdrproc_t outproc, char *out) {
	// WRITE MIRROR req
	msg_iod_t m;
	struct sockaddr_in *dasock = svc_getcaller(transp);
	u_int offset = argument.nfsproc_write_2_arg.offset;
	u_int datalen = argument.nfsproc_write_2_arg.data.data_len;
	int iod_no, ret;

	Dprintf(D_NFSP_DISPATCH, "forward_write_mirror : offset=%d datalen=%d", offset, datalen);

	// XXX BUG: we check if the request spans onto several BLOCKS:
	// for instance, read 8K at offset BLOCKSIZE-4K
	if ((offset % BLOCKSIZE) + datalen > BLOCKSIZE) {
		Dprintf(L_WARNING, "(offset[0x%8x]%%BLOCKSIZE[0x%8x]) + datalen[0x%8x] = %d > BLOCKSIZE",
			offset, BLOCKSIZE, datalen, (offset % BLOCKSIZE) + datalen);
		return -1;
	}

	m.magic = NFSP_MAGIC_MSG;
	m.msgtype = NFSP_WRITEMIRROR_REQ;

	m.client_addr = dasock->sin_addr.s_addr;
	m.client_port = dasock->sin_port;


	if (IPtoMAC(m.client_addr, m.client_mac_addr)) {
		Dprintf(L_WARNING, "forward_write_mirror: IPtoMAC failed");
		return -1;
	}

	m.server_addr = 0;		// iod will fill it
	m.server_port = htons(NFSport);

	m.writemirrorreq.xid = get_xid(transp);

	m.writemirrorreq.stripe_bs = BLOCKSIZE;
	m.writemirrorreq.stripe_w = viods_nb;
	
	m.writemirrorreq.offset = offset;
	m.writemirrorreq.datalen = datalen;

	m.writemirrorreq.inode = global_real_inode;
	m.writemirrorreq.seed = global_seed;

	memcpy(&m.writemirrorreq.attributes, &result.attrstat.attrstat_u.attributes,
		sizeof(result.attrstat.attrstat_u.attributes));

	Dprintf(D_NFSP_DISPATCH, "forward_write_mirror data_len=%d rev=%d",
			argument.nfsproc_write_2_arg.data.data_len,
			ntohl(argument.nfsproc_write_2_arg.data.data_len));

	memcpy(m.writemirrorreq.dataval,
		argument.nfsproc_write_2_arg.data.data_val, datalen);

	iod_no = find_iod(m.writemirrorreq.seed, m.writemirrorreq.offset);

	// fill idx and nextiods tab
	// TODO explain more of this idx stuff
	m.writemirrorreq.idx=get_all_iod(iod_no,m.writemirrorreq.nextiods);
	m.writemirrorreq.idx--;
	ret = sendto(iods_udp_fd, (char *)&m, MSG_IOD_WRITE_MIRROR_LEN(datalen), 0,
			get_iod(iod_no), sizeof(struct sockaddr_in));
	return ret;
}
#endif /* WRITEMIRROR_SUPPORT */


static int handle_remove(SVCXPRT *transp, xdrproc_t outproc, char *out) {
	Dprintf(D_NFSP_DISPATCH, "handle_remove");

	// We check if it is a "real" remove: if inode is 0 (invalid inode) then
	// we've nothing to do since the file is still hardlinked

	if (0 != global_real_inode) {
		int ret;
		remove_req_t req;

		req.magic = REMOVE_REQ_MAGIC;
		req.inode = global_real_inode;
		req.seed = global_seed;

		ret = write(unlinkpipe, &req, sizeof(remove_req_t));
		if (sizeof(remove_req_t) != ret) {
			if (-1 == ret) {
				Dprintf(L_WARNING, "remove request for i%x s%x in pipe=%d FAILED errno=%d",
				    req.inode, req.seed, unlinkpipe, errno);
			} else {
				Dprintf(L_WARNING, "remove request for i%x s%x in pipe=%d FAILED: wrote %d instead of %d",
				    req.inode, req.seed, unlinkpipe, ret, sizeof(remove_req_t));
			}
		} else {
			Dprintf(D_NFSP_DISPATCH, "remove request for i%x s%x succesfully written in pipe %d",
			     unlinkpipe, req.inode, req.seed);
		}
	}

	return svc_sendreply(transp, outproc,out);
}

// By using this function array we avoid post treatment testing & processing
typedef int (*svc_sendreply_t)( SVCXPRT *xprt, xdrproc_t outproc, char *out);
static svc_sendreply_t postfunctioncall[] = {
	svc_sendreply,		/* NULL */
	svc_sendreply,		/* GETATTR */
	svc_sendreply,		/* SETATTR */
	svc_sendreply,		/* ROOT */
	svc_sendreply,		/* LOOKUP */
	svc_sendreply,		/* READLINK */
	forward_read,		/* READ */
	svc_sendreply,		/* WRITECACHE */
#ifdef WRITEMIRROR_SUPPORT
#warning "WRITEMIRROR_SUPPORT ENABLED"
	forward_write_mirror,   /* WRITE MIRROR */
#else /* WRITEMIRROR_SUPPORT */
	forward_write,		/* WRITE */
#endif /* WRITEMIRROR_SUPPORT */
	svc_sendreply,		/* CREATE */
	handle_remove,		/* REMOVE */
	svc_sendreply,		/* RENAME */
	svc_sendreply,		/* LINK */
	svc_sendreply,		/* SYMLINK */
	svc_sendreply,		/* MKDIR */
	svc_sendreply,		/* RMDIR */
	svc_sendreply,		/* READDIR */
	svc_sendreply,		/* STATFS */
};

#ifdef CALL_PROFILING
#define PATH_PROFILE	"/tmp/nfspd.profile"

static struct timeval	rtimes[18] = {
	{0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
	{0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
	{0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}
};
static int		calls[18] = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
};
static struct timeval	t0, t1;
#endif


/*
 * The main dispatch routine.
 */
void
nfs_dispatch(struct svc_req *rqstp, SVCXPRT *transp)
{
	unsigned int proc_index = rqstp->rq_proc;
	struct dispatch_entry *dent;

	_rpcsvcdirty = 1;

	Dprintf(D_NFSP_DISPATCH, "---------------------------------------------");

	/* Reset our credentials to some sane default.
	 * Root privs will be needed in auth_fh/fh_find in order 
	 * to successfully stat() existing file handles */
	auth_override_uid(ROOT_UID);

	if (proc_index >= (sizeof(dtable) / sizeof(dtable[0]))) {
		svcerr_noproc(transp);
		goto done;
	}
	dent = &dtable[proc_index];

#ifdef CALL_PROFILING
	gettimeofday(&t0, NULL);
#endif

	/* Initialize our variables for determining the attributes of
	 * the file system in nfsd.c */
	nfsclient = NULL;

	memset(&argument, 0, dent->arg_size);
	if (!svc_getargs(transp, (xdrproc_t) dent->xdr_argument, &argument)) {
		svcerr_decode(transp);
		goto done;
	}

#ifdef VIODS_SUPPORT
	if (get_state() == NFSPD_DROPPING_MODE) {
		goto done;
	}
#endif /* VIODS_SUPPORT */

	/* Clear the result structure. */
	memset(&result, 0, dent->res_size);

	/* Log the call. The if() saves us a superfluous call to the debug
	 * functions, which is a big performance win.
	 */
	if (logging_enabled(D_CALL))
		log_call(rqstp, dent->name, dent->log_print(&argument));

	/* Do the function call itself. */
	nfs_dispatch_time = time(NULL);

	Dprintf(D_NFSP_DISPATCH, "Calling %s %s", dent->name, dent->log_print(&argument));

	result.nfsstat = (*dent->funct) (&argument, rqstp);
	Dprintf(D_CALL, "result: %d\n", result.nfsstat);

	// This array of functions contains a svc_sendreply call for every index
	// BUT for read, write and remove
	(postfunctioncall[proc_index])(transp, dent->xdr_result, (caddr_t) &result);

	if (!svc_freeargs(transp, (xdrproc_t) dent->xdr_argument, &argument)) {
		Dprintf(L_ERROR, "unable to free RPC arguments, exiting\n");
		exit(1);
	}
#ifdef CALL_PROFILING
	gettimeofday(&t1, NULL);

	if (t1.tv_usec < t0.tv_usec) {
		rtimes[proc_index].tv_sec += t1.tv_sec - t0.tv_sec - 1;
		rtimes[proc_index].tv_usec += 1000000 + t1.tv_usec - t0.tv_usec;
	} else {
		rtimes[proc_index].tv_sec += t1.tv_sec - t0.tv_sec;
		rtimes[proc_index].tv_usec += t1.tv_usec - t0.tv_usec;
	}
	calls[proc_index]++;
#endif

done:
	_rpcsvcdirty = 0;
	if (need_reinit) {
		reinitialize(0);
	}
	if (need_flush) {
		flush_cache(0);
	}
}

#ifdef CALL_PROFILING
void
dump_stats(int sig)
{
	FILE	*fp;
	int	i;

	if ((fp = fopen(PATH_PROFILE, "w")) == NULL) {
		Dprintf(L_WARNING, "unable to write profile data to %s\n",
				PATH_PROFILE);
		return;
	}

	for (i = 0; i < 18; i++) {
		float t;

		t = (float) rtimes[i].tv_sec +
		    (float) rtimes[i].tv_usec / 1000000.0;
		fprintf(fp, "%-20s\t%5d calls %8.4f sec avg\n",
				dtable[i].name, calls[i],
				(calls[i])?  t / calls[i] : 0);
		rtimes[i].tv_sec = rtimes[i].tv_usec = 0;
		calls[i] = 0;
	}

	fclose (fp);
}
#endif


/*
 * Functions for debugging output. This is still risky, because malformed
 * requests could overwrite our data segment.
 */
static char	printbuf[2048];

static char *
pr_void(void)
{
	return ("");
}

static char *
pr_nfs_fh(nfs_fh *argp)
{
	return (fh_pr(argp));
}

static char *
pr_sattrargs(sattrargs *argp)
{
	sprintf(printbuf, "fh:%s m:%0o u/g:%d/%d size:%d atime:%#x mtime:%#x",
		fh_pr(&argp->file), argp->attributes.mode,
		argp->attributes.uid, argp->attributes.gid,
		argp->attributes.size,
	 argp->attributes.atime.seconds, argp->attributes.mtime.seconds);
	return (printbuf);
}

static char *
pr_diropargs(diropargs *argp)
{
	sprintf(printbuf, "fh:%s n:%s", fh_pr(&(argp->dir)), argp->name);
	return (printbuf);
}

static char *
pr_readargs(readargs *argp)
{
	sprintf(printbuf, "%s: %d bytes at %d",
		fh_pr(&argp->file), argp->count, argp->offset);
	return (printbuf);
}

static char *
pr_writeargs(writeargs *argp)
{
	sprintf(printbuf, "%s: %d bytes at %d",
		fh_pr(&argp->file), argp->data.data_len, argp->offset);
	return (printbuf);
}

static char *
pr_createargs(createargs *argp)
{
	sprintf(printbuf, "fh:%s n:%s m:%0o u/g:%d/%d size:%d atime:%#x mtime:%#x",
		fh_pr(&argp->where.dir), argp->where.name,
		argp->attributes.mode, argp->attributes.uid,
		argp->attributes.gid, argp->attributes.size,
		argp->attributes.atime.seconds,
		argp->attributes.mtime.seconds);
	return (printbuf);
}

static char *
pr_renameargs(renameargs *argp)
{
	sprintf(printbuf, "fh:%s n:%s -> fh:%s n:%s",
		fh_pr(&argp->from.dir), argp->from.name,
		fh_pr(&argp->to.dir),   argp->to.name);
	return (printbuf);
}

static char *
pr_linkargs(linkargs *argp)
{
	sprintf(printbuf, "fh:%s -> fh:%s n:%s",
		fh_pr(&argp->from), fh_pr(&argp->to.dir), argp->to.name);
	return (printbuf);
}

static char *
pr_symlinkargs(symlinkargs *argp)
{
	return ("");
}

static char *
pr_readdirargs(readdirargs *argp)
{
	return (fh_pr(&argp->dir));
}
